#include <LocoNet.h>
#include <SPI.h>
#include <Ethernet.h>

#define TCP_PORT 1234

/* Modify these variables to match your network */
byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };
byte ip[] = { 192, 168, 0, 102 };
byte gateway[] = { 192, 168, 0, 1 };
byte subnet[] = { 255, 255, 255, 0 };

const char digitMap[] = "0123456789abcdef";
const char receiveString[] = "RECEIVE ";
static LnBuf LnTxBuffer;
static lnMsg *LnPacket;

/* LoconetOverTCP defaults to PORT 1024 */
EthernetServer server(TCP_PORT);
boolean gotATCPMessage = false;  // whether or not you got a message from the client yet


static void SendLoconetMessageViaTCP(lnMsg *LoconetPacket, uint8_t PacketSize);
static boolean ProcessTCPRxByte(uint8_t offset, uint8_t rxByte, LnBuf *LnTxBuffer);

void setup() {
  /* First initialize the LocoNet interface */
  LocoNet.init(7);

  /* Configure the serial port for 57600 baud */
  Serial.begin(57600);
  Serial.println("¡¡LoconetOverTCP buffer inicializado!!");

  /*Initialize a LocoNet packet buffer to buffer bytes from the PC */
  initLnBuf(&LnTxBuffer);

  /* initialise the ethernet device */
  Ethernet.begin(mac, ip, gateway, subnet);

  /*start listening for clients */
  server.begin();
}

void loop() {
  EthernetClient TcpClient = server.available();


  char RxByte = 0;
  static boolean IdenFound = false;
  static int IdenBytes = 0;
  int MsgIdx = 8;


  /* Check for any received loconet packets */
  LnPacket = LocoNet.receive();
  if (LnPacket) {
    /* If a packet has been received, get the length of the 
       loconet packet */
    uint8_t LnPacketSize = getLnMsgSize(LnPacket);


    /* Send the loconet packet out over TCP */
    SendLoconetMessageViaTCP(LnPacket, LnPacketSize);
  }


  /* Check for any received TCP packets */
  if (TcpClient) {
    if (!gotATCPMessage) {
      Serial.println("Hay un nuevo cliente");
      server.write("VERSION ArduinoLoconetEtherBuffer V0.2 built 17 November 2011\r\n");
      gotATCPMessage = true;
    }

    RxByte = TcpClient.read();

    switch (RxByte) {
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        ProcessTCPRxByte(0x30, (uint8_t)RxByte, &LnTxBuffer);
        break;
      case 'A':
      case 'B':
      case 'C':
      case 'F':
        ProcessTCPRxByte(0x37, (uint8_t)RxByte, &LnTxBuffer);
        break;

      case ' ':
        /* Do nothing */
        break;

      case 'S':
        IdenBytes = 1;
        if (IdenFound == true) {
          IdenFound = false;
        }
        break;

      case 'E':
        if (IdenFound == true)
          ProcessTCPRxByte(0x37, (uint8_t)RxByte, &LnTxBuffer);
        else
          IdenBytes++;
        break;

      case 'N':
        IdenBytes++;
        break;

      case 'D':
        if (IdenFound == true) {
          ProcessTCPRxByte(0x37, (uint8_t)RxByte, &LnTxBuffer);
        } else {
          if (IdenBytes == 3) {
            IdenFound = true;
          }
          IdenBytes = 0;
        }
        break;

      case -1:
      default:
        if (IdenFound == true) {
          server.write("SENT OK \r\n");
          IdenFound = false;
          IdenBytes = 0;
        }
        break;
    }
  } else {
    IdenFound = false;
    IdenBytes = 0;
  }
}



static boolean ProcessTCPRxByte(uint8_t offset, uint8_t rxByte, LnBuf *LnTxBuffer) {
  static boolean highNibble = true;
  static uint8_t loconetByte = 0;
  static lnMsg *LoconetPacket;

  if (highNibble) {
    loconetByte = (rxByte - offset) << 4;
    highNibble = false;
    return false;
  } else {
    loconetByte |= (rxByte - offset);
    highNibble = true;

    /* Add it to the buffer */

    addByteLnBuf(LnTxBuffer, loconetByte);

    /* Check to see if we have received a complete packet yet */
    LoconetPacket = recvLnMsg(LnTxBuffer);
    if (LoconetPacket) {
      /* Send the packet received via TCP to the LocoNet*/
      LocoNet.send(LoconetPacket);

      return true;
    }
    return false;
  }
}


static void SendLoconetMessageViaTCP(lnMsg *LoconetPacket, uint8_t PacketSize) {
  char TcpTxMsg[8 + (PacketSize * 3) + 2];
  uint8_t MsgIdx = 8;

  /* Initialise the the buffer that will be use to construct the message that is to 
           be transmitted over LoconetOverTCP*/
  memset(TcpTxMsg, 0, sizeof(TcpTxMsg));
  memcpy(TcpTxMsg, receiveString, sizeof(receiveString));


  // Send the received packet out byte by byte to the PC
  for (uint8_t Index = 0; Index < PacketSize; Index++) {
    /* Convert the high nibble of the current Loconet Packet byte to ASCII */
    TcpTxMsg[MsgIdx++] = (digitMap[(LoconetPacket->data[Index] >> 4) & 0xF]);

    /* Convert the low nibble of the current Loconet Packet byte to ASCII */
    TcpTxMsg[MsgIdx++] = (digitMap[(LoconetPacket->data[Index] & 0xF)]);

    /* Add a "space" between the current and next "loconet byte" */
    TcpTxMsg[MsgIdx++] = ' ';
  }
  /* Decrement the last MsgIdx, as there is no need for a space after the last byte */
  MsgIdx--;

  /* For some reason JMRI likes to have a "\r\n" at the end of the TCP message, though the
       loconetOverTCP standard states that an \r or an \n is accepetable */
  TcpTxMsg[MsgIdx++] = 0xD;
  TcpTxMsg[MsgIdx++] = 0xA;
  TcpTxMsg[MsgIdx] = 0;

  /* Send the message out over TCP */
  server.write(TcpTxMsg);
}
